import pytest

from pubtools.pulplib import Repository, ContainerImageRepository


def test_from_data_gives_container_image_repository():
    """Repository.from_data routes to ContainerImageRepository subclass when needed"""
    repo = Repository.from_data(
        {"id": "my-repo", "notes": {"_repo-type": "docker-repo"}}
    )
    assert isinstance(repo, ContainerImageRepository)


def test_default_registry_id():
    """registry_id defaults to repository ID"""
    assert ContainerImageRepository(id="foo").registry_id == "foo"


def test_registry_id_from_distributor():
    """registry_id is loaded from distributor when possible"""
    repo = Repository.from_data(
        {
            "id": "my-repo",
            "notes": {"_repo-type": "docker-repo"},
            "distributors": [
                {
                    "id": "docker_web_distributor_name_cli",
                    "distributor_type_id": "docker_distributor_web",
                    "config": {"repo-registry-id": "some/repo"},
                }
            ],
        }
    )

    assert repo.registry_id == "some/repo"


@pytest.mark.parametrize(
    "config", [{}, {"repo-registry-id": ""}, {"repo-registry-id": None}]
)
def test_default_registry_id_from_distributor(config):
    """default registry_id is used when it's not set in distributor or set to null/empty string"""
    repo = Repository.from_data(
        {
            "id": "my-repo",
            "notes": {"_repo-type": "docker-repo"},
            "distributors": [
                {
                    "id": "docker_web_distributor_name_cli",
                    "distributor_type_id": "docker_distributor_web",
                    "config": config,
                }
            ],
        }
    )

    assert repo.registry_id == "my-repo"



import time

from pubtools.pulplib import YumRepository, Distributor


def test_publish_cancel(fast_poller, requests_mocker, client, caplog):
    """Cancelling a publish future will cancel running Pulp task(s)."""

    repo = YumRepository(
        id="some-repo",
        distributors=(
            Distributor(id="yum_distributor", type_id="yum_distributor"),
            Distributor(id="cdn_distributor", type_id="cdn_distributor"),
        ),
    )
    repo.__dict__["_client"] = client

    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/repositories/some-repo/actions/publish/",
        [{"json": {"spawned_tasks": [{"task_id": "task1"}]}}],
    )

    task_search_url = "https://pulp.example.com/pulp/api/v2/tasks/search/"
    requests_mocker.post(
        task_search_url, [{"json": [{"task_id": "task1", "state": "running"}]}]
    )

    requests_mocker.delete("https://pulp.example.com/pulp/api/v2/tasks/task1/")

    # Start the publish
    publish_f = repo.publish()

    # Wait until we're sure poll thread has seen this task
    for _ in range(0, 1000):
        if (
            requests_mocker.last_request
            and requests_mocker.last_request.url == task_search_url
        ):
            break
        time.sleep(0.001)
    assert requests_mocker.last_request.url == task_search_url

    # We should be able to cancel it
    assert publish_f.cancel()

    # It should have cancelled the underlying Pulp task
    task_req = [
        r
        for r in requests_mocker.request_history
        if r.url == "https://pulp.example.com/pulp/api/v2/tasks/task1/"
    ]
    assert len(task_req) == 1
    assert task_req[0].method == "DELETE"



import sys
import os

import pytest

from pubtools.pulplib import (
    FakeController,
    YumRepository,
    ModulemdUnit,
    ModulemdDefaultsUnit,
    ModulemdDependency,
)


@pytest.mark.parametrize("use_file_object", [False, True])
def test_can_upload_units(use_file_object, data_path):
    """repo.upload_modules() succeeds with fake client and populates units."""
    modules_path = os.path.join(data_path, "sample-modules.yaml")

    controller = FakeController()

    controller.insert_repository(YumRepository(id="repo1"))

    client = controller.client
    repo1 = client.get_repository("repo1").result()

    to_upload = modules_path
    if use_file_object:
        to_upload = open(to_upload, "rb")

    upload_f = repo1.upload_modules(to_upload)

    # Upload should complete successfully.
    tasks = upload_f.result()

    # At least one task.
    assert tasks

    # Every task should have succeeded.
    for t in tasks:
        assert t.succeeded

    # If I now search for content in that repo, or content across all repos...
    units_in_repo = sorted(repo1.search_content().result(), key=repr)
    units_all = sorted(client.search_content().result(), key=repr)

    # They should be equal
    assert units_all == units_in_repo

    # And they should be this
    assert units_in_repo == [
        ModulemdDefaultsUnit(
            unit_id="e3e70682-c209-4cac-629f-6fbed82c07cd",
            name="ant",
            repo_id="repo1",
            # Note, this tests that 1.10 does not get coerced to 1.1,
            # as happened in some tools previously.
            stream="1.10",
            profiles={"1.10": ["default"]},
            content_type_id="modulemd_defaults",
            repository_memberships=["repo1"],
        ),
        ModulemdDefaultsUnit(
            unit_id="d4713d60-c8a7-0639-eb11-67b367a9c378",
            name="dwm",
            repo_id="repo1",
            stream=None,
            profiles={
                "6.0": ["default"],
                "6.1": ["default"],
                "6.2": ["default"],
                "latest": ["default"],
            },
            content_type_id="modulemd_defaults",
            repository_memberships=["repo1"],
        ),
        ModulemdUnit(
            unit_id="82e2e662-f728-b4fa-4248-5e3a0a5d2f34",
            name="avocado-vt",
            stream="82lts",
            version=3420210902113311,
            context="035be0ad",
            arch="x86_64",
            content_type_id="modulemd",
            repository_memberships=["repo1"],
            artifacts=[
                "avocado-vt-0:82.0-3.module_f34+12808+b491ffc8.src",
                "python3-avocado-vt-0:82.0-3.module_f34+12808+b491ffc8.noarch",
            ],
            profiles={
                "default": {
                    "description": "Common profile installing the avocado-vt plugin.",
                    "rpms": ["python3-avocado-vt"],
                }
            },
            dependencies=[ModulemdDependency(name="avocado", stream="82lts")],
        ),
        ModulemdUnit(
            unit_id="23a7711a-8133-2876-37eb-dcd9e87a1613",
            name="dwm",
            stream="6.0",
            version=3420210201213909,
            context="058368ca",
            arch="x86_64",
            content_type_id="modulemd",
            repository_memberships=["repo1"],
            artifacts=[
                "dwm-0:6.0-1.module_f34+11150+aec78cf8.src",
                "dwm-0:6.0-1.module_f34+11150+aec78cf8.x86_64",
                "dwm-debuginfo-0:6.0-1.module_f34+11150+aec78cf8.x86_64",
                "dwm-debugsource-0:6.0-1.module_f34+11150+aec78cf8.x86_64",
                "dwm-user-0:6.0-1.module_f34+11150+aec78cf8.x86_64",
            ],
            profiles={
                "default": {
                    "description": "The minimal, distribution-compiled dwm binary.",
                    "rpms": ["dwm"],
                },
                "user": {
                    "description": "Includes distribution-compiled dwm as well as a helper script to apply user patches and configuration, dwm-user.",
                    "rpms": ["dwm", "dwm-user"],
                },
            },
            dependencies=[],
        ),
    ]



from datetime import datetime
import logging

import pytest
from more_executors.futures import f_return
from mock import patch, Mock, call
from requests.exceptions import HTTPError

from pubtools.pulplib._impl.model.repository.repo_lock import RepoLock, LockClaim
from pubtools.pulplib import Client, FakeController, Repository


@pytest.fixture
def mock_client():
    client = Client("https://pulp.example.com/")
    client._request_executor.submit = Mock()
    client._do_request = "do_request_func"
    return client


def pulp_get_response(key_values):
    return f_return({"notes": key_values})


def pulp_put_response():
    return f_return({})


@patch("pubtools.pulplib._impl.model.repository.repo_lock.now")
@patch("pubtools.pulplib._impl.model.repository.repo_lock.random")
def test_repo_lock(random_mock, datetime_mock, mock_client):
    datetime_mock.side_effect = [
        datetime.fromtimestamp(1000.0),
        datetime.fromtimestamp(1060.0),
        datetime.fromtimestamp(1120.0),
        datetime.fromtimestamp(1120.0),
        datetime.fromtimestamp(1120.0),
        datetime.fromtimestamp(1180.0),
    ]
    random_mock.choices = Mock(return_value="1234")
    # Pulp API Calls
    # 1. Get the current locks to check if three's stale ones. None this time.
    # 2. Put our lock in the repo notes.
    # 3. Get the current locks to check if it's our turn to run. (not yet)
    # 3. Get the current locks again, it's our turn this time.
    # 4. Delete the lock from the repo notes.
    mock_client._request_executor.submit.side_effect = [
        pulp_get_response({}),  # 1
        pulp_put_response(),  # 2
        pulp_get_response(
            {
                "pulplib-lock-claim-1233": '{"id": "1233", "context": "testing repo lock", "created": 900.0, "valid_from": 910.0, "expires": 1900.0}',
                "pulplib-lock-claim-1234": '{"id": "1234", "context": "testing repo lock", "created": 1000.0, "valid_from": 1010.0, "expires": 2000.0}',
            }
        ),  # 3
        pulp_get_response(
            {
                "pulplib-lock-claim-1234": '{"id": "1234", "context": "testing repo lock", "created": 1000.0, "valid_from": 1010.0, "expires": 2000.0}'
            }
        ),  # 4
        pulp_put_response(),  # 5
    ]
    expected_endpoint = "https://pulp.example.com/pulp/api/v2/repositories/test-repo/"
    expected_mock_calls = [
        call("do_request_func", url=expected_endpoint, method="GET"),
        call(
            "do_request_func",
            url=expected_endpoint,
            method="PUT",
            json={
                "delta": {
                    "notes": {
                        "pulplib-lock-claim-1234": '{"id": "1234", "context": "testing repo lock", "created": 1000.0, "valid_from": 1010.0, "expires": 2000.0}'
                    }
                }
            },
        ),
        call("do_request_func", url=expected_endpoint, method="GET"),
        call("do_request_func", url=expected_endpoint, method="GET"),
        call(
            "do_request_func",
            url=expected_endpoint,
            method="PUT",
            json={"delta": {"notes": {"pulplib-lock-claim-1234": None}}},
        ),
    ]
    lock = RepoLock(
        "test-repo",
        mock_client,
        "testing repo lock",
        1000,
    )

    # Runs __enter__ and __exit__ in RepoLock
    with lock:
        pass

    mock_client._request_executor.submit.assert_has_calls(expected_mock_calls)
    assert mock_client._request_executor.submit.call_count == 5


@patch("pubtools.pulplib._impl.model.repository.repo_lock.now")
@patch("pubtools.pulplib._impl.model.repository.repo_lock.random")
def test_repo_lock_fake_client(random_mock, datetime_mock):
    datetime_mock.side_effect = [
        datetime.fromtimestamp(1000.0),
        datetime.fromtimestamp(1060.0),
        datetime.fromtimestamp(1120.0),
        datetime.fromtimestamp(1180.0),
    ]
    random_mock.choices = Mock(return_value="1234")
    ctrl = FakeController()
    ctrl.insert_repository(Repository(id="test-repo"))
    lock = ctrl.new_client().get_repository("test-repo").lock("test")

    # Runs __enter__ and __exit__ in RepoLock
    with lock:
        pass

    lock_history = ctrl.repo_lock_history
    assert len(lock_history) == 2
    assert lock_history[0].repository == "test-repo"
    assert lock_history[0].action == "lock"
    assert lock_history[1].repository == "test-repo"
    assert lock_history[1].action == "unlock"


@patch("pubtools.pulplib._impl.model.repository.repo_lock.now")
def test_remove_expired_locks(datetime_mock, mock_client):
    datetime_mock.return_value = datetime.fromtimestamp(3000.0)

    mock_client._request_executor.submit.side_effect = [
        pulp_get_response(
            {
                "pulplib-lock-claim-1234": '{"id": "1234", "context": "testing repo lock", "created": 1000.0, "valid_from": 1010.0, "expires": 2000.0}',
                "pulplib-lock-claim-1235": '{"id": "1235", "context": "testing repo lock", "created": 1000.0, "valid_from": 1010.0, "expires": 2000.0}',
            }
        ),
        pulp_put_response(),
    ]
    expected_endpoint = "https://pulp.example.com/pulp/api/v2/repositories/test-repo/"
    expected_mock_calls = [
        call("do_request_func", url=expected_endpoint, method="GET"),
        call(
            "do_request_func",
            url=expected_endpoint,
            method="PUT",
            json={
                "delta": {
                    "notes": {
                        "pulplib-lock-claim-1234": None,
                        "pulplib-lock-claim-1235": None,
                    }
                }
            },
        ),
    ]
    lock = RepoLock(
        "test-repo",
        mock_client,
        "testing repo lock",
        1000,
    )

    lock.remove_expired_locks()
    mock_client._request_executor.submit.assert_has_calls(expected_mock_calls)


@patch("pubtools.pulplib._impl.model.repository.repo_lock.now")
def test_remove_expired_locks_fake_client(datetime_mock):
    datetime_mock.return_value = datetime.fromtimestamp(3000.0)
    ctrl = FakeController()
    ctrl.insert_repository(Repository(id="test-repo"))
    client = ctrl.new_client()
    client._update_repo_lock_data(
        "test-repo",
        {
            "pulplib-lock-claim-1234": '{"id": "1234", "context": "testing repo lock", "created": 1000.0, "valid_from": 1010.0, "expires": 2000.0}'
        },
    )
    client._update_repo_lock_data(
        "test-repo",
        {
            "pulplib-lock-claim-1235": '{"id": "1235", "context": "testing repo lock", "created": 1000.0, "valid_from": 1010.0, "expires": 2000.0}'
        },
    )

    client.get_repository("test-repo").lock("test").remove_expired_locks()

    lock_history = ctrl.repo_lock_history
    assert len(lock_history) == 3
    assert lock_history[0].repository == "test-repo"
    assert lock_history[0].action == "lock"
    assert lock_history[1].repository == "test-repo"
    assert lock_history[1].action == "lock"
    assert lock_history[2].repository == "test-repo"
    assert lock_history[2].action == "multi-unlock"


@patch("pubtools.pulplib._impl.model.repository.repo_lock.now")
def test_remove_expired_locks_error(datetime_mock, mock_client, caplog):
    # We may cause an error if we try to delete a lock which has already been
    # deleted. This may happen if another task runs clean up at the same time.
    mock_response = Mock()
    mock_response.json.return_value = {
        "exception": ["KeyError: 'pulplib-lock-claim-1234'\n"]
    }
    mock_client._update_repo_lock_data = Mock(
        side_effect=HTTPError(response=mock_response)
    )
    datetime_mock.return_value = datetime.fromtimestamp(3000.0)

    mock_client._request_executor.submit.side_effect = [
        pulp_get_response(
            {
                "pulplib-lock-claim-1234": '{"id": "1234", "context": "testing repo lock", "created": 1000.0, "valid_from": 1010.0, "expires": 2000.0}',
                "pulplib-lock-claim-1235": '{"id": "1235", "context": "testing repo lock", "created": 1000.0, "valid_from": 1010.0, "expires": 2000.0}',
            }
        ),
    ]

    lock = RepoLock(
        "test-repo",
        mock_client,
        "testing repo lock",
        1000,
    )

    lock.remove_expired_locks()

    assert (
        "An error occurred while trying to delete an expired lock. The locks have already been deleted"
        in caplog.messages
    )


def test_exception_handling_on_lock_delete(mock_client):
    mock_response = Mock()
    mock_response.json.return_value = {
        "exception": ["KeyError: 'pulplib-lock-claim-1234'\n"]
    }
    mock_client._update_repo_lock_data = Mock(
        side_effect=HTTPError(response=mock_response)
    )
    expected_mock_calls = [
        call("test-repo", {"pulplib-lock-claim-1234": None}, await_result=True)
    ]
    lock = RepoLock(
        "test-repo",
        mock_client,
        "testing repo lock",
        1000,
    )
    lock._lock_claim = LockClaim.from_json_data(
        {
            "id": "1234",
            "context": "test",
            "created": 1000.0,
            "valid_from": 1010.0,
            "expires": 2000.0,
        }
    )

    lock.delete_lock_claim()

    mock_client._update_repo_lock_data.assert_has_calls(expected_mock_calls)


@patch("pubtools.pulplib._impl.model.repository.repo_lock.now")
@patch("pubtools.pulplib._impl.model.repository.repo_lock.random")
def test_repo_log_error_on_stale_lock(random_mock, datetime_mock, caplog):
    # If we've held the lock for too long, there should be an error reported.
    datetime_mock.side_effect = [
        datetime.fromtimestamp(1000.0),
        datetime.fromtimestamp(1060.0),
        datetime.fromtimestamp(1120.0),
        datetime.fromtimestamp(5000.0),
        datetime.fromtimestamp(5000.0),
    ]
    random_mock.choices = Mock(return_value="1234")
    ctrl = FakeController()
    ctrl.insert_repository(Repository(id="test-repo"))
    lock = ctrl.new_client().get_repository("test-repo").lock("test")

    # Runs __enter__ and __exit__ in RepoLock
    with lock:
        pass

    # Should have logged an error.
    assert (
        "Client requested lock on repo test-repo for 1800 seconds but was held for 4000 seconds (test)"
        in caplog.messages
    )



import datetime
import pytest
import attr

from pubtools.pulplib import FakeController, FileUnit, FileRepository


def test_update_no_id():
    controller = FakeController()
    client = controller.client

    # Try to update something with no ID; it should fail immediately
    # (no future) as we can't even try to update without an ID.
    with pytest.raises(ValueError) as excinfo:
        client.update_content(
            FileUnit(
                path="x",
                size=0,
                sha256sum="e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855",
            )
        )

    # It should tell us why
    assert "unit_id missing on call to update_content()" in str(excinfo.value)


def test_update_nonexistent():
    controller = FakeController()
    client = controller.client

    # Try to update something which doesn't exist.
    update_f = client.update_content(
        FileUnit(
            unit_id="this-unit-is-missing",
            path="x",
            size=0,
            sha256sum="e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855",
        )
    )

    # It should fail, async.
    assert "unit not found: this-unit-is-missing" in str(update_f.exception())


def test_can_update_content(tmpdir):
    controller = FakeController()

    controller.insert_repository(FileRepository(id="repo1"))

    client = controller.client
    repo1 = client.get_repository("repo1").result()

    somefile = tmpdir.join("some-file.txt")
    somefile.write(b"there is some binary data:\x00\x01\x02")

    upload_f = repo1.upload_file(
        str(somefile),
        description="My great file",
        version="1.0",
        cdn_path="/foo/bar.txt",
    )

    # The future should resolve successfully
    tasks = upload_f.result()

    # The task should be successful.
    assert tasks[0].succeeded

    # File should now be in repo.
    units_in_repo = list(repo1.search_content())
    assert len(units_in_repo) == 1
    unit = units_in_repo[0]

    # Sanity check we got the right thing.
    assert unit.path == "some-file.txt"

    # Now let's try updating fields.
    new_unit = attr.evolve(
        unit,
        size=1000,
        description="My greater file",
        display_order=10,
        cdn_published=datetime.datetime(2021, 12, 6, 11, 19, 0),
    )

    update_f = client.update_content(new_unit)

    # It should succeed.
    update_f.result()

    # Now get the unit again.
    units_in_repo = list(repo1.search_content())
    assert len(units_in_repo) == 1
    unit_after_update = units_in_repo[0]

    # The mutable fields on the unit after update should be as expected:
    assert unit_after_update.description == "My greater file"
    assert unit_after_update.cdn_published == datetime.datetime(2021, 12, 6, 11, 19, 0)
    assert unit_after_update.display_order == 10
    # This one didn't change since it wasn't evolved
    assert unit_after_update.cdn_path == "/foo/bar.txt"
    assert unit_after_update.version == "1.0"
    # This one didn't change (even though we tried) because it's not mutable
    assert unit_after_update.size == 29



import pytest

from pubtools.pulplib import (
    FakeController,
    RpmUnit,
    YumRepository,
    CopyOptions,
)


@pytest.fixture
def controller():
    return FakeController()


def test_clients_share_state(controller):
    """Multiple clients created by the same controller share state"""
    src = YumRepository(id="src-repo")
    dest = YumRepository(id="dest-repo")
    controller.insert_repository(src)
    controller.insert_repository(dest)

    src_units = [
        RpmUnit(name="bash", version="4.0", release="1", arch="x86_64"),
        RpmUnit(name="bash", version="4.0", release="2", arch="x86_64"),
        RpmUnit(name="bash", version="4.1", release="3", arch="x86_64"),
        RpmUnit(name="glibc", version="5.0", release="1", arch="x86_64"),
    ]
    controller.insert_units(src, src_units)

    client1 = controller.new_client()
    client2 = controller.new_client()

    # The two clients should not be the same object
    assert client1 is not client2

    # Repos are initially detached, re-fetch them via client
    src = client1.get_repository(src.id).result()
    dest = client1.get_repository(dest.id).result()

    # Do a copy via client1
    client1.copy_content(
        src, dest, options=CopyOptions(require_signed_rpms=False)
    ).result()

    # Then search for content via client2
    found_units = list(client2.search_content())

    # It should be able to see the outcome of the copy done via client1;
    # i.e. repository_memberships contains both repos after the copy
    assert sorted(found_units, key=repr) == [
        RpmUnit(
            name="bash",
            version="4.0",
            release="1",
            arch="x86_64",
            repository_memberships=["src-repo", "dest-repo"],
            unit_id="e3e70682-c209-4cac-629f-6fbed82c07cd",
        ),
        RpmUnit(
            name="bash",
            version="4.0",
            release="2",
            arch="x86_64",
            repository_memberships=["src-repo", "dest-repo"],
            unit_id="82e2e662-f728-b4fa-4248-5e3a0a5d2f34",
        ),
        RpmUnit(
            name="bash",
            version="4.1",
            release="3",
            arch="x86_64",
            repository_memberships=["src-repo", "dest-repo"],
            unit_id="d4713d60-c8a7-0639-eb11-67b367a9c378",
        ),
        RpmUnit(
            name="glibc",
            version="5.0",
            release="1",
            arch="x86_64",
            repository_memberships=["src-repo", "dest-repo"],
            unit_id="23a7711a-8133-2876-37eb-dcd9e87a1613",
        ),
    ]



from io import BytesIO
from xml.parsers.expat import ExpatError

import pytest

from pubtools.pulplib import YumRepository


def test_upload_comps_xml(client, requests_mocker):
    """A client can upload expected units from within a comps.xml."""

    repo_id = "repo1"
    repo = YumRepository(id=repo_id)
    repo.__dict__["_client"] = client

    # Some XML to upload.
    # This is fairly minimal as this test focuses on upload functionality.
    # There are other tests exercising the XML parser more thoroughly.
    xml = BytesIO(
        b"""
        <comps>
            <group>
                <id>3d-printing</id>
                <name>3D Printing</name>
                <name xml:lang="af">3D-drukwerk</name>
                <description>3D printing software</description>
                <packagelist>
                    <packagereq type="default">admesh</packagereq>
                    <packagereq>blender</packagereq>
                </packagelist>
            </group>

            <category>
                <id>kde-desktop-environment</id>
                <name>KDE Desktop</name>
                <display_order>10</display_order>
                <grouplist>
                    <groupid>kde-office</groupid>
                    <groupid>kde-telepathy</groupid>
                </grouplist>
            </category>


            <environment>
                <id>basic-desktop-environment</id>
                <name>Basic Desktop</name>
                <name xml:lang="af">Basiese werkskerm</name>
                <description>X Window System with a choice of window manager.</description>
                <grouplist>
                    <groupid>networkmanager-submodules</groupid>
                    <groupid>standard</groupid>
                </grouplist>
                <optionlist>
                    <groupid default="true">xmonad</groupid>
                    <groupid>xmonad-mate</groupid>
                </optionlist>
            </environment>

            <langpacks>
                <match install="stardict-dic-%s" name="stardict"/>
                <match install="tkgate-%s" name="tkgate"/>
            </langpacks>
        </comps>
    """
    )

    # Set up the requests it'll do:
    #
    # It should unassociate previous units
    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/repositories/repo1/actions/unassociate/",
        json={"spawned_tasks": [{"task_id": "remove-task"}]},
    )

    # It'll search for the status of tasks.
    # Note: to avoid an overly complex mock we just let every task search return all
    # tasks rather than differentiating between remove and upload tasks
    upload_count = 4
    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/tasks/search/",
        json=[{"task_id": "remove-task", "state": "finished"}]
        + [
            {"task_id": "upload-task-%d" % i, "state": "finished"}
            for i in range(0, upload_count)
        ],
    )

    # It will request various uploads
    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/content/uploads/",
        [{"json": {"upload_id": "upload-%d" % i}} for i in range(0, upload_count)],
    )

    # It'll do an import against the uploads
    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/repositories/repo1/actions/import_upload/",
        [
            {"json": {"spawned_tasks": [{"task_id": "upload-task-%d" % i}]}}
            for i in range(0, upload_count)
        ],
    )

    # It'll clean up the uploads
    for i in range(0, upload_count):
        requests_mocker.delete(
            "https://pulp.example.com/pulp/api/v2/content/uploads/upload-%d/" % i
        )

    # Mocker setup is complete.

    # It should start the upload OK
    upload_f = repo.upload_comps_xml(xml)

    # I can await the result
    tasks = upload_f.result()

    # There should be various tasks returned from the API
    assert sorted([t.id for t in tasks]) == [
        "upload-task-0",
        "upload-task-1",
        "upload-task-2",
        "upload-task-3",
    ]

    # Each task should be successful
    for t in tasks:
        assert t.completed
        assert t.succeeded

    # Gather info on exactly what the client imported
    imported_units = []
    for req in requests_mocker.request_history:
        if req.url.endswith("/import_upload/"):
            imported_units.append(req.json()["unit_metadata"])

    imported_units.sort(key=lambda u: u["_content_type_id"])

    # Did we import exactly the expected units?
    assert imported_units == [
        {
            "_content_type_id": "package_category",
            "id": "kde-desktop-environment",
            "name": "KDE Desktop",
            "display_order": 10,
            "packagegroupids": ["kde-office", "kde-telepathy"],
            "repo_id": "repo1",
        },
        {
            "_content_type_id": "package_environment",
            "id": "basic-desktop-environment",
            "name": "Basic Desktop",
            "translated_name": {"af": "Basiese werkskerm"},
            "description": "X Window System with a choice of window manager.",
            "group_ids": ["networkmanager-submodules", "standard"],
            "options": [
                {"group": "xmonad", "default": True},
                {"group": "xmonad-mate", "default": False},
            ],
            "repo_id": "repo1",
        },
        {
            "_content_type_id": "package_group",
            "id": "3d-printing",
            "name": "3D Printing",
            "translated_name": {"af": "3D-drukwerk"},
            "description": "3D printing software",
            "default_package_names": ["admesh"],
            "mandatory_package_names": ["blender"],
            "repo_id": "repo1",
        },
        {
            "_content_type_id": "package_langpacks",
            "matches": [
                {"install": "stardict-dic-%s", "name": "stardict"},
                {"install": "tkgate-%s", "name": "tkgate"},
            ],
            "repo_id": "repo1",
        },
    ]


def test_upload_empty_comps_xml(client, requests_mocker):
    """A client can upload an empty comps.xml, which merely removes existing units."""

    repo_id = "repo1"
    repo = YumRepository(id=repo_id)
    repo.__dict__["_client"] = client

    xml = BytesIO(b"<comps/>")

    # Set up the requests it'll do:
    #
    # It should unassociate previous units
    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/repositories/repo1/actions/unassociate/",
        json={"spawned_tasks": [{"task_id": "remove-task"}]},
    )

    # It'll search for the status of tasks.
    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/tasks/search/",
        json=[{"task_id": "remove-task", "state": "finished"}],
    )

    # And... that's it.

    # It should start the upload OK
    upload_f = repo.upload_comps_xml(xml)

    # I can await the result
    tasks = upload_f.result()

    # It should just return the removal task
    assert [t.id for t in tasks] == ["remove-task"]
    assert tasks[0].completed
    assert tasks[0].succeeded

    # Check exactly the removed types
    assert requests_mocker.request_history[0].json() == {
        "criteria": {
            "type_ids": [
                "package_group",
                "package_category",
                "package_environment",
                "package_langpacks",
            ]
        }
    }


def test_bad_comps_xml(client, requests_mocker, tmpdir):
    """A client requested to upload an invalid comps.xml will raise immediately
    without making any requests to Pulp."""

    repo_id = "repo1"
    repo = YumRepository(id=repo_id)
    repo.__dict__["_client"] = client

    comps_xml = tmpdir.join("comps.xml")
    comps_xml.write(b"Oops not valid comps")

    # It should raise
    with pytest.raises(ExpatError):
        repo.upload_comps_xml(str(comps_xml))

    # As we did not register anything in requests_mocker, we've already
    # implicitly tested that no requests happened. But just to be clear...
    assert not requests_mocker.request_history



from pubtools.pulplib import YumRepository, ErratumUnit, ErratumReference


def test_upload_erratum(client, requests_mocker):
    """An ErratumUnit can be uploaded to a repository using the client."""

    repo_id = "repo1"
    repo = YumRepository(id=repo_id)
    repo.__dict__["_client"] = client

    unit = ErratumUnit(
        id="RHBA-1234:56",
        summary="A great advisory",
        version="1",
        status="final",
        updated="some updated time",
        issued="some issued time",
        description="testing upload",
        pushcount="507",
        from_="noreply@example.com",
        reboot_suggested=False,
        rights="copyright bob",
        content_types=["rpm", "module"],
        references=[
            ErratumReference(
                title="a link",
                href="https://example.com/test-advisory",
                type="self",
                id="self-id",
            )
        ],
        pkglist=[],
        container_list=None,
    )

    # Set up the requests it'll do:
    #
    # It will request an upload
    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/content/uploads/",
        json={"upload_id": "my-upload-123"},
    )

    # It'll do an import against that upload
    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/repositories/repo1/actions/import_upload/",
        json={"spawned_tasks": [{"task_id": "task123"}, {"task_id": "task234"}]},
    )

    # It'll search for status of the import tasks
    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/tasks/search/",
        json=[
            {"task_id": "task123", "state": "finished"},
            {"task_id": "task234", "state": "finished"},
        ],
    )

    # It'll clean up the upload
    requests_mocker.delete(
        "https://pulp.example.com/pulp/api/v2/content/uploads/my-upload-123/"
    )

    # Mocker setup is complete.

    # It should start the upload OK
    upload_f = repo.upload_erratum(unit)

    # I can await the result
    tasks = upload_f.result()

    # There should be the two tasks returned from the API
    assert sorted([t.id for t in tasks]) == ["task123", "task234"]

    # Each task should be successful
    for t in tasks:
        assert t.completed
        assert t.succeeded

    # What did the client actually try to import?
    import_request = requests_mocker.request_history[1]

    assert import_request.json() == {
        "upload_id": "my-upload-123",
        "unit_type_id": "erratum",
        "unit_key": {"id": "RHBA-1234:56"},
        "unit_metadata": {
            # This should contain an accurate serialization of all fields
            # into the form required by Pulp.
            "id": "RHBA-1234:56",
            "version": "1",
            "status": "final",
            "updated": "some updated time",
            "issued": "some issued time",
            "description": "testing upload",
            "pushcount": "507",
            "reboot_suggested": False,
            "from": "noreply@example.com",
            "rights": "copyright bob",
            "summary": "A great advisory",
            "pulp_user_metadata": {
                "content_types": ["rpm", "module"],
                "container_list": [],
            },
            "references": [
                {
                    "href": "https://example.com/test-advisory",
                    "id": "self-id",
                    "title": "a link",
                    "type": "self",
                }
            ],
            "pkglist": [],
            # Note also that fields not explicitly set will be serialized
            # as None (see doc string on upload_erratum).
            "title": None,
            "severity": None,
            "release": None,
            "type": None,
            "solution": None,
        },
    }



from pubtools.pulplib import FakeController, Repository


def test_create_repository():
    """Client.create_repository() with fake client adds new repositories to controller."""
    controller = FakeController()

    client = controller.client
    repo_1 = client.create_repository(Repository(id="repo1"))
    repo_2 = client.create_repository(Repository(id="repo2"))

    # adding already existing repository has no effect
    _ = client.create_repository(Repository(id="repo1"))
    # The change should be reflected in the controller,
    # with two repositories present
    assert controller.repositories == [repo_1.result(), repo_2.result()]



import pytest
import logging
import requests

from pubtools.pulplib import Repository, YumRepository, DetachedException, YumImporter


def test_from_data_gives_yum_repository():
    """Repository.from_data maps to YumRepository subclass when needed"""
    repo = Repository.from_data({"id": "my-repo", "notes": {"_repo-type": "rpm-repo"}})
    assert isinstance(repo, YumRepository)


def test_default_mutable_urls():
    """mutable_urls has expected default value"""
    assert YumRepository(id="foo").mutable_urls == ["repodata/repomd.xml"]


def test_can_hash():
    """a default YumRepository is hashable"""
    repo = YumRepository(id="foo")
    reposet = set()
    reposet.add(repo)
    assert repo in reposet


def test_from_data_relative_url():
    """relative_url is initialized from distributors when possible"""
    repo = Repository.from_data(
        {
            "id": "my-repo",
            "notes": {"_repo-type": "rpm-repo"},
            "distributors": [
                {
                    "id": "yum_distributor",
                    "distributor_type_id": "yum_distributor",
                    "config": {"relative_url": "some/publish/path"},
                }
            ],
        }
    )

    assert repo.relative_url == "some/publish/path"


def test_from_data_skip_rsync_repodata():
    """skip_rsync_repodata is initialized from distributors when possible"""
    repo = Repository.from_data(
        {
            "id": "my-repo",
            "notes": {"_repo-type": "rpm-repo"},
            "distributors": [
                {
                    "id": "cdn_distributor",
                    "distributor_type_id": "rpm_rsync_distributor",
                    "config": {"skip_repodata": True},
                }
            ],
        }
    )
    assert repo.skip_rsync_repodata


def test_from_data_importer():
    repo = Repository.from_data(
        {
            "id": "my-repo",
            "notes": {"_repo-type": "rpm-repo"},
            "importers": [
                {
                    "importer_type_id": "yum_importer",
                    "config": {"foo": "bar"},
                }
            ],
        }
    )

    assert repo.importer.type_id == "yum_importer"
    assert repo.importer.config == {"foo": "bar"}


def test_populate_attrs():
    """test populate attributes are correctly parsed from repo notes"""
    repo = Repository.from_data(
        {
            "id": "my-repo",
            "notes": {
                "_repo-type": "rpm-repo",
                "content_set": "fake_content_set",
                "include_in_download_service": "True",
                "include_in_download_service_preview": "True",
                "population_sources": ["populate_repo1", "populate_repo2"],
                "provisioning": True,
                "ubi_config_version": "fake_ubi_config_version",
                "ubi_population": True,
            },
            "distributors": [],
        }
    )
    assert repo.content_set == "fake_content_set"
    assert repo.include_in_download_service
    assert repo.include_in_download_service_preview
    assert repo.population_sources == ["populate_repo1", "populate_repo2"]
    assert repo.provisioning
    assert repo.ubi_config_version == "fake_ubi_config_version"
    assert repo.ubi_population


def test_productid_attrs():
    """All attributes relating to productid are correctly parsed from repo notes."""

    repo = Repository.from_data(
        {
            "id": "my-repo",
            "notes": {
                "_repo-type": "rpm-repo",
                "arch": "x86_64",
                "product_versions": '["1.4", "1.100", "1.2"]',
                "platform_full_version": "whatever",
                "eng_product": "123",
            },
            "distributors": [],
        }
    )

    assert repo.arch == "x86_64"
    assert repo.eng_product_id == 123
    assert repo.platform_full_version == "whatever"

    # Note the version-aware sorting: 1.100 is larger than 1.4
    assert repo.product_versions == ["1.2", "1.4", "1.100"]


def test_product_versions_unusual_attrs():
    """Odd values in product_versions are tolerated."""

    repo = Repository.from_data(
        {
            "id": "my-repo",
            "notes": {
                "_repo-type": "rpm-repo",
                "product_versions": '["1.4", 234, "1.100", "not numeric"]',
            },
            "distributors": [],
        }
    )

    assert repo.product_versions == ["1.100", "1.4", "234", "not numeric"]


def test_related_repositories(client, requests_mocker):
    """test Repository.get_*_repository returns expected objects"""

    repo_binary_test = YumRepository(id="repo_binary", relative_url="some/repo/os")
    repo_binary_test.__dict__["_client"] = client

    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/distributors/search/",
        [
            {
                "json": [
                    {
                        "id": "yum_distributor",
                        "distributor_type_id": "yum_distributor",
                        "repo_id": "repo_debug",
                        "config": {"relative_url": "some/repo/debug"},
                    }
                ]
            },
            {
                "json": [
                    {
                        "id": "yum_distributor",
                        "distributor_type_id": "yum_distributor",
                        "repo_id": "repo_source",
                        "config": {"relative_url": "some/repo/SRPMS"},
                    }
                ]
            },
        ],
    )

    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/repositories/search/",
        [{"json": [{"id": "repo_debug"}]}, {"json": [{"id": "repo_source"}]}],
    )

    # Request for binary repo should return identical object
    assert repo_binary_test is repo_binary_test.get_binary_repository().result()
    assert repo_binary_test.get_binary_repository().id == "repo_binary"
    # Requests for debug and source repositories return correct repositories
    assert repo_binary_test.get_debug_repository().id == "repo_debug"
    assert repo_binary_test.get_source_repository().id == "repo_source"


def test_related_repositories_not_found(client, requests_mocker):
    """test Repository.get_*_repository returns Future[None] if repository is not found"""

    repo_binary_test = YumRepository(id="repo_binary", relative_url="some/repo/os")
    repo_binary_test.__dict__["_client"] = client

    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/distributors/search/", json=[]
    )

    repo = repo_binary_test.get_source_repository()
    assert repo.result() is None


def test_related_repositories_detached_client():
    repo_binary_test = YumRepository(id="repo_binary", relative_url="some/repo/os")
    repo_binary_test.__dict__["_client"] = None

    with pytest.raises(DetachedException):
        repo_binary_test.get_binary_repository()


def test_create_repository(client, requests_mocker):
    repo = YumRepository(id="yum_repo_new")

    # create request
    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/repositories/",
        json={},
    )

    # repo search request
    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/repositories/search/",
        json=[
            {
                "id": "yum_repo_new",
                "notes": {"_repo-type": "rpm-repo"},
                "importers": [
                    {
                        "importer_type_id": "yum_importer",
                        "config": {},
                    }
                ],
            }
        ],
    )

    out = client.create_repository(repo)
    # check return value of create_repository() call
    assert out.result() == repo

    hist = requests_mocker.request_history
    # there should be exactly 2 requests sent - create and search
    assert len(hist) == 2

    create_query = hist[0].json()
    # check id of repository sent in request body
    assert create_query["id"] == "yum_repo_new"
    # check importer data sent in request body that
    # are automatically added for yum repository
    assert create_query["importer_type_id"] == "yum_importer"
    assert create_query["importer_config"] == {}

    # check the search request for created repo
    search_query = hist[1].json()
    assert search_query == {
        "criteria": {
            "skip": 0,
            "limit": 2000,
            "filters": {"id": {"$eq": "yum_repo_new"}},
        },
        "distributors": True,
        "importers": True,
    }


def test_create_repository_already_exists(client, requests_mocker, caplog):
    repo = YumRepository(id="yum_repo_existing")

    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/repositories/",
        status_code=409,
        text="Conflict 409 status",
    )

    # repo search request
    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/repositories/search/",
        json=[
            {
                "id": "yum_repo_existing",
                "notes": {"_repo-type": "rpm-repo"},
                "importers": [
                    {
                        "importer_type_id": "yum_importer",
                        "config": {},
                    }
                ],
            }
        ],
    )

    with caplog.at_level(logging.WARNING):
        out = client.create_repository(repo)
        # check return value of create_repository() call
        assert out.result() == repo

        hist = requests_mocker.request_history
        # there should be 2 request sent - attempt to create and search for repo,
        # 409 status is never retried
        assert len(hist) == 2

        query = hist[0].json()
        # check id of repository sent in request body
        assert query["id"] == "yum_repo_existing"
        # check importer data sent in request body that
        # are automatically added for yum repository
        assert query["importer_type_id"] == "yum_importer"
        assert query["importer_config"] == {}
        # check logged information about existing repository
        assert "Repository yum_repo_existing already exists" in caplog.text

        # check the search request for existing repo
        search_query = hist[1].json()
        assert search_query == {
            "criteria": {
                "skip": 0,
                "limit": 2000,
                "filters": {"id": {"$eq": "yum_repo_existing"}},
            },
            "distributors": True,
            "importers": True,
        }


def test_create_repository_wrong_data(client, requests_mocker, caplog):
    repo = YumRepository(
        id="yum_repo_existing", importer=YumImporter(config={"new": "value"})
    )

    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/repositories/",
        status_code=409,
        text="Conflict 409 status",
    )

    # repo search request
    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/repositories/search/",
        json=[
            {
                "id": "yum_repo_existing",
                "notes": {"_repo-type": "rpm-repo"},
                "importers": [
                    {
                        "importer_type_id": "yum_importer",
                        "config": {"current": "value"},
                    }
                ],
            }
        ],
    )
    with caplog.at_level(logging.WARNING):
        with pytest.raises(AssertionError):
            client.create_repository(repo).result()

        for text in (
            "Repository yum_repo_existing already exists",
            "Repository yum_repo_existing contains wrong importer config",
            "Repository yum_repo_existing exists on server and contains unexpected values",
        ):
            assert text in caplog.text


def test_create_repository_raises_exception(client, requests_mocker):
    repo = YumRepository(id="yum_repo")

    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/repositories/",
        status_code=400,
        text="Client error 400",
    )

    with pytest.raises(requests.exceptions.HTTPError):
        client.create_repository(repo).result()



import pytest

from pubtools.pulplib import (
    FakeController,
    YumRepository,
    RpmUnit,
    ModulemdUnit,
    Criteria,
)


def test_can_remove_empty():
    """repo.remove() succeeds with empty result if repo has no content."""
    controller = FakeController()
    client = controller.client

    repo = YumRepository(id="repo1")
    controller.insert_repository(repo)

    remove_tasks = client.get_repository("repo1").remove_content()

    assert len(remove_tasks) == 1
    task = remove_tasks[0]

    # It should have completed successfully
    assert task.completed
    assert task.succeeded

    # But should not have removed any units
    assert not task.units


def test_limited_remove_content():
    """repo.remove() can remove a limited number of units."""
    controller = FakeController()
    client = controller.client

    rpm_units = [
        RpmUnit(name="gliba", version="1.0", release="1", arch="x86_64"),
        RpmUnit(name="glibb", version="1.0", release="1", arch="x86_64"),
        RpmUnit(name="glibc", version="1.0", release="1", arch="x86_64"),
        RpmUnit(name="glibd", version="1.0", release="1", arch="x86_64"),
    ]

    repo = YumRepository(id="repo1")
    controller.insert_repository(repo)
    controller.insert_units(repo, rpm_units)

    remove_rpms = client.get_repository("repo1").remove_content(
        type_ids=["rpm"], limit=3
    )

    assert len(remove_rpms) == 1
    task = remove_rpms[0]

    # It should have completed successfully
    assert task.completed
    assert task.succeeded

    # It should have removed (only) RPM units
    assert len(task.units) == 3


def test_can_remove_content():
    """repo.remove() succeeds and removes expected units inserted via controller."""
    controller = FakeController()
    client = controller.client

    rpm_units = [
        RpmUnit(name="bash", version="4.0", release="1", arch="x86_64"),
        RpmUnit(name="glibc", version="5.0", release="1", arch="x86_64"),
    ]
    modulemd_units = [
        ModulemdUnit(
            name="module1", stream="s1", version=1234, context="a1b2", arch="x86_64"
        ),
        ModulemdUnit(
            name="module1", stream="s1", version=1235, context="a1b2", arch="x86_64"
        ),
    ]
    units = rpm_units + modulemd_units

    repo = YumRepository(id="repo1")
    controller.insert_repository(repo)
    controller.insert_units(repo, units)

    remove_rpms = client.get_repository("repo1").remove_content(type_ids=["rpm"])

    assert len(remove_rpms) == 1
    task = remove_rpms[0]

    # It should have completed successfully
    assert task.completed
    assert task.succeeded

    # It should have removed (only) RPM units
    assert sorted(task.units) == sorted(rpm_units)

    # Now if we ask to remove same content again...
    remove_rpms = client.get_repository("repo1").remove_content(type_ids=["rpm"])

    assert len(remove_rpms) == 1
    task = remove_rpms[0]

    # It should have completed successfully, but no RPMs to remove
    assert task.completed
    assert task.succeeded
    assert not task.units

    # It should still be possible to remove other content
    remove_all = client.get_repository("repo1").remove_content()

    assert len(remove_all) == 1
    task = remove_all[0]

    # It should have completed successfully, and removed the modulemds
    assert task.completed
    assert task.succeeded
    assert sorted(task.units) == sorted(modulemd_units)


def test_remove_deleted_repo_fails():
    """repo.remove() fails if repository doesn't exist."""
    controller = FakeController()
    client = controller.client

    repo = YumRepository(id="repo1")
    controller.insert_repository(repo)

    # Get two references to the same repo
    repo_handle1 = client.get_repository("repo1")
    repo_handle2 = client.get_repository("repo1")

    # Use one of them to delete the repo
    repo_handle1.delete().result()

    # Now if I try to use the other one to remove content...
    remove = repo_handle2.remove_content()

    # It should fail
    assert "Repository id=repo1 not found" in str(remove.exception())


def test_remove_filtered_content():
    """repo.remove_content() succeeds and removes expected units inserted via controller."""
    controller = FakeController()
    client = controller.client

    rpm_units = [
        RpmUnit(name="bash", version="4.0", release="1", arch="x86_64"),
        RpmUnit(name="glibc", version="5.0", release="1", arch="x86_64"),
    ]
    modulemd_units = [
        ModulemdUnit(
            name="module1", stream="s1", version=1234, context="a1b2", arch="x86_64"
        ),
        ModulemdUnit(
            name="module1", stream="s1", version=1235, context="a1b2", arch="x86_64"
        ),
    ]
    units = rpm_units + modulemd_units

    repo = YumRepository(id="repo1")
    controller.insert_repository(repo)
    controller.insert_units(repo, units)

    criteria = Criteria.and_(
        Criteria.with_unit_type(RpmUnit), Criteria.with_field("name", "bash")
    )
    remove_rpms = client.get_repository("repo1").remove_content(criteria=criteria)

    assert len(remove_rpms) == 1
    task = remove_rpms[0]

    # It should have completed successfully
    assert task.completed
    assert task.succeeded

    # It should have removed the specific rpm
    assert len(task.units) == 1
    assert task.units[0] == sorted(rpm_units)[0]

    # Try removing a module with incorrect type_ids
    criteria = Criteria.and_(
        Criteria.with_unit_type(RpmUnit), Criteria.with_field("name", "module1")
    )
    remove_rpms = client.get_repository("repo1").remove_content(criteria=criteria)

    assert len(remove_rpms) == 1
    task = remove_rpms[0]

    # It should have completed successfully
    assert task.completed
    assert task.succeeded

    # Nothing's removed as criteria isn't fulfilled
    assert len(task.units) == 0

    # Removing module with correct type_ids
    criteria = Criteria.and_(
        Criteria.with_unit_type(ModulemdUnit), Criteria.with_field("name", "module1")
    )
    remove_rpms = client.get_repository("repo1").remove_content(criteria=criteria)

    assert len(remove_rpms) == 1
    task = remove_rpms[0]

    # It should have completed successfully
    assert task.completed
    assert task.succeeded

    # It should have removed both the modules as they
    # match the criteria
    assert len(task.units) == 2
    assert sorted(task.units) == sorted(modulemd_units)


def test_remove_fails_without_type_id():
    """repo.remove_content() fails when a criteria is provided without unit type"""
    controller = FakeController()
    client = controller.client

    rpm_units = [
        RpmUnit(name="bash", version="4.0", release="1", arch="x86_64"),
        RpmUnit(name="glibc", version="5.0", release="1", arch="x86_64"),
    ]

    repo = YumRepository(id="repo1")
    controller.insert_repository(repo)
    controller.insert_units(repo, rpm_units)

    criteria = Criteria.and_(Criteria.with_field("name", "bash"))

    with pytest.raises(ValueError):
        client.get_repository("repo1").remove_content(criteria=criteria)



import logging
import pytest

from pubtools.pulplib import (
    Repository,
    YumRepository,
    Task,
    Distributor,
    DetachedException,
    PublishOptions,
    TaskFailedException,
    RpmUnit,
    FileUnit,
    ModulemdUnit,
    YumRepoMetadataFileUnit,
    Unit,
    Criteria,
    Matcher,
)


def test_detached():
    """remove_content raises if called on a detached repo"""
    with pytest.raises(DetachedException):
        Repository(id="some-repo").remove_content()


def test_remove_no_type_ids(fast_poller, requests_mocker, client):
    """Remove succeeds when given no type_ids."""

    repo = Repository(id="some-repo")
    repo.__dict__["_client"] = client

    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/repositories/some-repo/actions/unassociate/",
        [{"json": {"spawned_tasks": [{"task_id": "task1"}, {"task_id": "task2"}]}}],
    )

    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/tasks/search/",
        [
            {
                "json": [
                    {"task_id": "task1", "state": "finished"},
                    {"task_id": "task2", "state": "skipped"},
                ]
            }
        ],
    )

    assert repo.remove_content().result() == [
        Task(id="task1", completed=True, succeeded=True),
        Task(id="task2", completed=True, succeeded=True),
    ]


def test_remove_with_type_ids(fast_poller, requests_mocker, client):
    """Remove succeeds when given specific type_ids."""

    repo = Repository(id="some-repo")
    repo.__dict__["_client"] = client

    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/repositories/some-repo/actions/unassociate/",
        [{"json": {"spawned_tasks": [{"task_id": "task1"}]}}],
    )

    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/tasks/search/",
        [{"json": [{"task_id": "task1", "state": "finished"}]}],
    )

    assert repo.remove_content(type_ids=["type1", "type2"]).result() == [
        Task(id="task1", completed=True, succeeded=True)
    ]

    # It should have passed those type_ids to Pulp
    req = requests_mocker.request_history
    assert (
        req[0].url
        == "https://pulp.example.com/pulp/api/v2/repositories/some-repo/actions/unassociate/"
    )
    assert req[0].json() == {"criteria": {"type_ids": ["type1", "type2"]}}


def test_remove_loads_units(fast_poller, requests_mocker, client):
    """Remove returns unit info loaded from units_successful."""

    repo = Repository(id="some-repo")
    repo.__dict__["_client"] = client

    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/repositories/some-repo/actions/unassociate/",
        [{"json": {"spawned_tasks": [{"task_id": "task1"}]}}],
    )

    unit_data = [
        {
            "type_id": "iso",
            "unit_key": {"name": "hello.txt", "size": 23, "checksum": "a" * 64},
        },
        {
            "type_id": "rpm",
            "unit_key": {
                "name": "bash",
                "epoch": "0",
                "version": "4.0",
                "release": "1",
                "arch": "x86_64",
            },
        },
        {
            "type_id": "modulemd",
            "unit_key": {
                "name": "module",
                "stream": "s1",
                "version": 1234,
                "context": "a1b2c3",
                "arch": "s390x",
            },
        },
        {
            "type_id": "yum_repo_metadata_file",
            "unit_key": {"data_type": "productid", "repo_id": "some-repo"},
        },
        {"type_id": "bizarre_type", "unit_key": {"whatever": "data"}},
    ]

    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/tasks/search/",
        [
            {
                "json": [
                    {
                        "task_id": "task1",
                        "state": "finished",
                        "result": {"units_successful": unit_data},
                    }
                ]
            }
        ],
    )

    tasks = repo.remove_content().result()

    # It should return one task
    assert len(tasks) == 1
    task = tasks[0]

    # It should be the expected successful task
    assert task.id == "task1"
    assert task.completed
    assert task.succeeded

    # It should have loaded expected units from the units_successful dict
    assert set(task.units) == set(
        [
            FileUnit(path="hello.txt", size=23, sha256sum="a" * 64),
            RpmUnit(
                name="bash",
                epoch="0",
                version="4.0",
                release="1",
                arch="x86_64",
                sourcerpm=None,
            ),
            ModulemdUnit(
                name="module", stream="s1", version=1234, context="a1b2c3", arch="s390x"
            ),
            YumRepoMetadataFileUnit(
                data_type="productid", content_type_id="yum_repo_metadata_file"
            ),
            Unit(content_type_id="bizarre_type"),
        ]
    )


def test_remove_with_criteria(fast_poller, requests_mocker, client):
    """Remove succeeds when given a critria/filter for removal"""
    repo = Repository(id="some-repo")
    repo.__dict__["_client"] = client

    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/repositories/some-repo/actions/unassociate/",
        [
            {"json": {"spawned_tasks": [{"task_id": "task1"}]}},
            {"json": {"spawned_tasks": [{"task_id": "task2"}]}},
        ],
    )

    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/tasks/search/",
        [
            {"json": [{"task_id": "task1", "state": "finished"}]},
            {"json": [{"task_id": "task2", "state": "finished"}]},
        ],
    )
    criteria = Criteria.and_(
        Criteria.with_unit_type(RpmUnit),
        Criteria.with_field("filename", "some.rpm"),
        Criteria.with_field("signing_key", Matcher.in_(["key123"])),
    )

    assert repo.remove_content(criteria=criteria).result() == [
        Task(id="task1", completed=True, succeeded=True)
    ]

    # It should have passed the criteria to Pulp
    req = requests_mocker.request_history
    assert (
        req[0].url
        == "https://pulp.example.com/pulp/api/v2/repositories/some-repo/actions/unassociate/"
    )
    assert req[0].json() == {
        "criteria": {
            "filters": {
                "unit": {
                    "$and": [
                        {"filename": {"$eq": "some.rpm"}},
                        {"signing_key": {"$in": ["key123"]}},
                    ]
                }
            },
            "type_ids": ["rpm", "srpm"],
        }
    }

    # Providing both criteria and type_ids
    assert repo.remove_content(
        criteria=criteria, type_ids=["type1", "type2"]
    ).result() == [Task(id="task2", completed=True, succeeded=True)]

    # It should have passed only the critera to Pulp and ignore type_ids as kwarg
    req = requests_mocker.request_history
    assert (
        req[0].url
        == "https://pulp.example.com/pulp/api/v2/repositories/some-repo/actions/unassociate/"
    )
    assert req[0].json() == {
        "criteria": {
            "filters": {
                "unit": {
                    "$and": [
                        {"filename": {"$eq": "some.rpm"}},
                        {"signing_key": {"$in": ["key123"]}},
                    ]
                }
            },
            "type_ids": ["rpm", "srpm"],
        }
    }


def test_remove_with_limit(fast_poller, requests_mocker, client):
    """Remove succeeds when given a critria/filter for removal"""
    repo = Repository(id="some-repo")
    repo.__dict__["_client"] = client

    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/repositories/some-repo/actions/unassociate/",
        [
            {"json": {"spawned_tasks": [{"task_id": "task1"}]}},
        ],
    )

    requests_mocker.post(
        "https://pulp.example.com/pulp/api/v2/tasks/search/",
        [
            {"json": [{"task_id": "task1", "state": "finished"}]},
        ],
    )

    assert repo.remove_content(type_ids=["type1", "type2"], limit=1).result() == [
        Task(id="task1", completed=True, succeeded=True)
    ]

    # It should have included the limit in the post request
    assert requests_mocker.request_history[0].json() == {
        "criteria": {
            "type_ids": ["type1", "type2"],
            "limit": 1,
        }
    }


def test_remove_fail_without_type_id(fast_poller, client):
    """Remove fails when a critria is provided without unit type"""
    repo = Repository(id="some-repo")
    repo.__dict__["_client"] = client

    criteria = Criteria.and_(
        Criteria.with_field("filename", "some.rpm"),
        Criteria.with_field("signing_key", Matcher.in_(["key123"])),
    )

    with pytest.raises(ValueError):
        repo.remove_content(criteria=criteria)



import sys
import os

import pytest

from pubtools.pulplib import FakeController, YumRepoMetadataFileUnit, YumRepository


@pytest.mark.parametrize("use_file_object", [False, True])
def test_can_upload_units(use_file_object, tmpdir):
    """repo.upload_metadata() succeeds with fake client and populates units."""

    testfile = tmpdir.join("testfile")
    testfile.write(b"some bytes")

    controller = FakeController()

    controller.insert_repository(YumRepository(id="repo1"))

    client = controller.client
    repo1 = client.get_repository("repo1").result()

    to_upload = str(testfile)
    if use_file_object:
        to_upload = open(to_upload, "rb")

    upload_f = repo1.upload_metadata(to_upload, "my-metadata-type")

    # Upload should complete successfully.
    tasks = upload_f.result()

    # At least one task.
    assert tasks

    # Every task should have succeeded.
    for t in tasks:
        assert t.succeeded

    # If I now search for content in that repo, or content across all repos...
    units_in_repo = sorted(repo1.search_content().result(), key=lambda u: u.sha256sum)
    units_all = sorted(client.search_content().result(), key=lambda u: u.sha256sum)

    # They should be equal
    assert units_all == units_in_repo

    # And they should be this
    assert units_in_repo == [
        YumRepoMetadataFileUnit(
            data_type="my-metadata-type",
            sha256sum="0d22cdcc10e6d049dbe1af5123d50873fdfc1a4f58306e58cb6241be9472014d",
            content_type_id="yum_repo_metadata_file",
            repository_memberships=["repo1"],
            unit_id="e3e70682-c209-4cac-629f-6fbed82c07cd",
        )
    ]


def test_overwrite(tmpdir):
    """repo.upload_metadata() multiple times with same data type overwrites old data."""

    testfile1 = tmpdir.join("testfile1")
    testfile1.write(b"some bytes")
    testfile2 = tmpdir.join("testfile2")
    testfile2.write(b"other bytes")

    controller = FakeController()

    controller.insert_repository(YumRepository(id="repo1"))

    client = controller.client
    repo1 = client.get_repository("repo1").result()

    upload_f = repo1.upload_metadata(str(testfile1), "mdtype1")

    # Upload should complete successfully.
    upload_f.result()

    # It should have uploaded the first file
    assert list(repo1.search_content()) == [
        YumRepoMetadataFileUnit(
            data_type="mdtype1",
            sha256sum="0d22cdcc10e6d049dbe1af5123d50873fdfc1a4f58306e58cb6241be9472014d",
            content_type_id="yum_repo_metadata_file",
            repository_memberships=["repo1"],
            unit_id="e3e70682-c209-4cac-629f-6fbed82c07cd",
        )
    ]

    # Now upload different content with same type
    upload_f = repo1.upload_metadata(str(testfile2), "mdtype1")

    # Upload should complete successfully.
    upload_f.result()

    # There should still just be one unit, but with the updated content
    assert list(repo1.search_content()) == [
        YumRepoMetadataFileUnit(
            data_type="mdtype1",
            sha256sum="a3ead5eedad5df82318c51685dbc1c147a36d1ff8584fc82de6b08d0bf63a795",
            content_type_id="yum_repo_metadata_file",
            repository_memberships=["repo1"],
            unit_id="e3e70682-c209-4cac-629f-6fbed82c07cd",
        )
    ]

    # Uploading same content as a *different* type should be fine.
    upload_f = repo1.upload_metadata(str(testfile2), "mdtype2")

    # Upload should complete successfully.
    upload_f.result()

    # Now the same content is available as two types.
    assert sorted(repo1.search_content(), key=lambda u: u.data_type) == [
        YumRepoMetadataFileUnit(
            data_type="mdtype1",
            sha256sum="a3ead5eedad5df82318c51685dbc1c147a36d1ff8584fc82de6b08d0bf63a795",
            content_type_id="yum_repo_metadata_file",
            repository_memberships=["repo1"],
            unit_id="e3e70682-c209-4cac-629f-6fbed82c07cd",
        ),
        YumRepoMetadataFileUnit(
            data_type="mdtype2",
            sha256sum="a3ead5eedad5df82318c51685dbc1c147a36d1ff8584fc82de6b08d0bf63a795",
            content_type_id="yum_repo_metadata_file",
            repository_memberships=["repo1"],
            unit_id="e6f4590b-9a16-4106-cf6a-659eb4862b21",
        ),
    ]



import pytest

from mock import Mock

from pubtools.pulplib._impl.compat_attr import validators

from pubtools.pulplib._impl.model.validate import NamedMappingValidator


def test_named_mapping_validator_extra_keys():
    validator = NamedMappingValidator(
        mapping={"foo": validators.instance_of(str)}, type_=dict
    )
    with pytest.raises(ValueError):
        validator(None, Mock(name="attr"), {"foo": "1", "bar": "2"})


def test_named_mapping_validator_missing_keys():
    validator = NamedMappingValidator(
        mapping={
            "foo": validators.instance_of(str),
            "bar": validators.instance_of(str),
        },
        type_=dict,
    )
    with pytest.raises(ValueError):
        validator(None, Mock(name="attr"), {"foo": "1"})



import os

import pytest

from pubtools.pulplib import ErratumUnit

from pubtools.pulplib import InvalidDataException


def test_container_list_invalid(data_path):
    """Provide invalid structure object to container_list attribute."""

    with open(
        os.path.join(data_path, "sample-erratum-invalid-container-list1.json"), "rt"
    ) as f:
        erratum_data = f.read()

    with pytest.raises(InvalidDataException):
        ErratumUnit.from_data(erratum_data)
